import 'dart:io' show Directory, File, FileMode, IOSink;

import 'package:meta/meta.dart' show required;
import 'package:jetlog/jetlog.dart' show Handler, Record;
import 'package:jetlog/formatters.dart' show Formatter;

class FileHandler extends Handler {
  FileHandler(
      {@required Formatter formatter,
      @required this.filePath,
      @required this.maxSize})
      : assert(maxSize != null && !maxSize.isNegative),
        assert(filePath != null),
        _formatter = formatter,
        _currentSize = 0;

  final Formatter _formatter;
  File _file;
  IOSink _sink;

  int _currentSize;

  final Uri filePath;
  final int maxSize;

  String get _dirPath => filePath.toFilePath();

  String get _fileName => filePath.toFilePath();

  String get _backupFileName {
    return '';
  }

  Future<void> _close() => _sink.close();

  Future<void> _openExistingOrNewFile() async {
    _file = File(_fileName);

    // Sync is typically faster than asynchronous one.
    if (!_file.existsSync()) {
      return _openNewFile();
    }

    _sink = _file.openWrite(mode: FileMode.writeOnlyAppend);
    _currentSize = _file.lengthSync();
  }

  Future<void> _openNewFile() async {
    await Directory(_dirPath).create(recursive: true);

    await _file?.rename(_backupFileName);

    _file = File(_fileName);
    _sink = _file.openWrite(mode: FileMode.writeOnly);
    _currentSize = 0;
  }

  Future<void> _rotateFile() async {
    await _close();
    await _openNewFile();
  }

  @override
  void handle(Record record) {
    final message = _formatter.call(record);
    final size = message.length;

    if (size > maxSize) {
      throw Exception('Exceeded maximum file size!');
    }

    if (_file == null) {
      _openExistingOrNewFile();
    }

    if (_currentSize + size > maxSize) {
      _rotateFile();
    }

    _sink.add(message);
    _currentSize += size;
  }

  @override
  Future<void> close() async {
    await super.close();

    return _close();
  }
}
